********************************************************************
*** Name.....: CLASSFINDER.PRG
*** Author...: Andy Kramek & Marcia Akins
*** Date.....: 06/02/2005
*** Notice...: Copyright (c) 2005 Tightline Computers, Inc
*** Compiler.: Visual FoxPro 09.00.0000.2412 for Windows 
*** Function.: Crawl through a directory tree checks all SCX/PRG for usage of classes
********************************************************************
LPARAMETERS tcRootDir
*** If no directory passed, assume current directory
IF VARTYPE(tcRootDir) # "C" OR EMPTY( tcRootDir )
  tcRootDir = FULLPATH( CURDIR())
ENDIF
*** Make sure that the starting location exists
IF NOT DIRECTORY( tcRootDir, 1 )
  MESSAGEBOX( "Must pass a valid directory to Class Finder", 16, "Bad Call" )
  RETURN
ENDIF

*** All this is just so that we can DO the program and create the object
RELEASE oFinder
PUBLIC oFinder
oFinder = CREATEOBJECT( 'classfinder', tcRootDir )
SET DATASESSION TO oFinder.DataSessionID
oFinder.FindClasses()

DEFINE CLASS classfinder AS Session
  cRootDir = ''
  cHomeDir = ''
  oDirs    = NULL
  nLevels  = 0

  ********************************************************************
  *** [E] FINDCLASSES: Crawl through the directory tree looking for class usage
  ********************************************************************
  FUNCTION FindClasses( tcRootDir )
  LOCAL llRetVal, lcCurDir
    WITH This
      *** Did we get anything passed?
      IF VARTYPE(tcRootDir) # "C" OR EMPTY( tcRootDir )
        *** No, so use whatever was set up as the default
        tcRootDir = .cRootDir
      ENDIF
      *** Make this the current default directory and add it to the collection
      SET DEFAULT TO (tcRootDir)
      lcCurDir = FULLPATH( CURDIR())
      *** Clear the collection first and reset the level counter
      .oDirs.Remove( -1 )
      .nLevels = 1
      .oDirs.Add( lcCurDir, "DIR001-" + lcCurDir )
      *** Build the list of sub-directories
      .GetDirs( lcCurDir )

      *** Now we need to traverse each directory
      FOR lnCnt = 1 TO .oDirs.Count
        *** Get the directory name
        lcDir = .oDirs[lnCnt]
        *** Check SCX/VCX files
        .CheckSCX( lcDir )
      
        *** Check PRGs
        .CheckPrg( lcDir )
      NEXT

      *** Return Final Status    
      RETURN llRetVal
    ENDWITH
  ENDFUNC && FindClasses
  
  ********************************************************************
  *** [P] CHECKPRG: Check PRG files and get all classes
  ********************************************************************
  PROTECTED FUNCTION CheckPrg( tcSchDir )
  LOCAL ARRAY laPrg[1]
  LOCAL llRetVal, lcCurDir, lnPrg, lnCnt, lcFile
    WITH This
      *** Save directory and change to target
      lcCurDir = FULLPATH( CURDIR() )
      SET DEFAULT TO (tcSchDir )
      lnPrg = ADIR( laPrg, '*.prg' )
      IF lnPrg > 0
        FOR lnCnt = 1 TO lnPrg
          lcFile = laPrg[ lnCnt, 1 ]
          .ChkProgCode( lcFile )
        NEXT
      ENDIF

      *** Restore settings
      SET DEFAULT TO (lcCurDir)
       
      *** Return Final Status    
      RETURN llRetVal
    ENDWITH
  ENDFUNC && CheckPrg

  ********************************************************************
  *** [P] CHKPROGCODE: Open the specified PRG and get class info
  ********************************************************************
  PROTECTED FUNCTION ChkProgCode( tcFile )
  LOCAL ARRAY laLines[1]
  LOCAL lcCode, lnLines, lcLoc, lcFile, lnCnt
    WITH This
      TRY
        *** Get the contents into a string
        lcCode = FILETOSTR( tcFile )
        lnLines = ALINES( laLines, lcCode, 1 )
        IF lnLines < 1
          *** Empty!!!
          RETURN
        ENDIF

        *** Get the file name and location
        lcLoc = FULLPATH( CURDIR() )

        *** Process each line of code
        FOR lnCnt = 1 TO lnLines
          lcLine = LOWER( ALLTRIM( laLines[lnCnt] ))
          IF LEFT( lcLIne, 1 ) = "*"
            *** It's a comment
            LOOP
          ENDIF

          *** If we have the word object in the line
          IF 'object' $ lcLine
            *** Parse out the Class information 
            loRetObj = .GetNewClass( lcLine )
          
            *** Did we actually get a class back
            IF NOT EMPTY( loRetObj.cClass )
              *** There still may be a proc named "createnewobject" or something....
              INSERT INTO cur_Classes VALUES ( lcLoc, tcFile, loRetObj.cClass, loRetObj.cLib, ;
                                               loRetObj.cBase, loRetObj.cObjName, "" )
            ENDIF
          ENDIF  && 'object' $ lcLine
        NEXT
        
      CATCH TO loerr
        MESSAGEBOX( loErr.Message, 16, "Problem with " + tcFile )
     
      ENDTRY 
           
      *** Return Final Status    
      RETURN 
    ENDWITH
  ENDFUNC && ChkProgCode

  ********************************************************************
  *** [P] CHECKSCX: Check SCX files and get all classes
  ********************************************************************
  PROTECTED FUNCTION CheckSCX( tcSchDir )
  LOCAL ARRAY laScx[1]
  LOCAL llRetVal, lcCurDir, lnScx, lnCnt, lcFile
    WITH This
      *** Save directory and change to target
      lcCurDir = FULLPATH( CURDIR() )
      SET DEFAULT TO (tcSchDir )

      *** Check for Forms
      lnScx = ADIR( laScx, '*.scx' )
      IF lnScx > 0
        FOR lnCnt = 1 TO lnScx
          lcFile = laScx[ lnCnt, 1 ]
          .GetSCXClass( lcFile )
        NEXT
      ENDIF

      *** Check for Class Libraries
      lnScx = ADIR( laScx, '*.vcx' )
      IF lnScx > 0
        FOR lnCnt = 1 TO lnScx
          lcFile = laScx[ lnCnt, 1 ]
          .GetSCXClass( lcFile )
        NEXT
      ENDIF

      *** Restore settings
      SET DEFAULT TO (lcCurDir)
       
      *** Return Final Status    
      RETURN llRetVal
    ENDWITH
  ENDFUNC && CheckSCX
  
  ********************************************************************
  *** [P] GETSCXCLASS: Open the specified SCX as table and get class info
  ********************************************************************
  PROTECTED FUNCTION GetScxClass( tcFile )
  LOCAL llRetVal, lcLoc, lcFile, lcClass, lcLib, lcBase, lcObj, loData, lcParent
    WITH This
      TRY
        *** Open the SCX
        USE (tcFile) IN 0 ALIAS scefile
        SELECT scefile
        *** Get the file name and location
        lcLoc = JUSTPATH( DBF())
        lcFile = JUSTFNAME( DBF())
        GO TOP
        *** Make sure this is really a VFP SCX file
        SCATTER NAME loData MEMO
        IF PEMSTATUS( loData, 'class', 5 )
          SCAN
            *** Check for explicit references first
            IF NOT EMPTY( ALLTRIM( scefile.class ))
              *** Is this something other than a baseclass
              IF JUSTEXT( tcFile ) == "SCX"
                IF LOWER( ALLTRIM( scefile.class )) == LOWER( ALLTRIM( scefile.baseclass ))
                  *** This is a base class - ignore it
                  LOOP
                ENDIF
              ENDIF
              *** This is a class we are interested in
              lcClass = LOWER( ALLTRIM( scefile.class ))
              lcLib   = LOWER( ALLTRIM( scefile.classloc ))
              lcBase  = LOWER( ALLTRIM( scefile.baseclass ))
              lcObj   = LOWER( ALLTRIM( scefile.objname ))
              lcParent = LOWER( ALLTRIM( scefile.parent ) )
              *** So add it to the result cursor
              INSERT INTO cur_Classes VALUES ( lcLoc, lcFile, lcClass, lcLib, lcBase, lcObj, lcParent )
            ENDIF
            
            *** Now try for implicit ones (in the Methods memo field)
            IF 'object' $ LOWER( methods )
              *** May be an implicit declaration so call the method to check for it
              .ChkMethodCode( ALLTRIM( LOWER( methods)), lcLoc, lcFile )
            ENDIF
          ENDSCAN  
        ENDIF
      
      CATCH TO loerr
        MESSAGEBOX( loErr.Message, 16, "Problem with " + tcFile )
     
      ENDTRY 
           
      *** Close the SCX
      IF USED( 'scefile' )
        USE IN sceFile
      ENDIF
      
      *** Return Final Status    
      RETURN 
    ENDWITH
  ENDFUNC && GetScxClass

  ********************************************************************
  *** [P] CHKMETHODCODE: Check the method code for object declaration
  ********************************************************************
  PROTECTED FUNCTION ChkMethodCode( tcMethodCode, tcLoc, tcFile )
  LOCAL ARRAY laCode[1]
  LOCAL llRetVal, lnLines, lnCnt, lcLine, lcClass, lcLib
    WITH This
      *** Get the code, omitting blank lines
      lnLines = ALINES( laCode, tcMethodCode, 5 )
      FOR lnCnt = 1 TO lnLines
        *** Format the line with no spaces/quotes
        lcLine = ALLTRIM( CHRTRAN( laCode[ lnCnt ], "['" + CHR(9) + CHR(34) + " ]", '' ))
        IF LEFT( lcLine, 1 ) = "*"
          *** This is a comment - ignore it
          LOOP
        ENDIF

        *** If we have the word object in the line
        IF 'object' $ lcLine
          *** Parse out the Class information 
          loRetObj = .GetNewClass( lcLine )
            
          *** Did we actually get a class back
          IF NOT EMPTY( loRetObj.cClass )
            INSERT INTO cur_Classes VALUES ( tcLoc, tcFile, loRetObj.cClass, loRetObj.cLib, ;
                                             loRetObj.cBase, loRetObj.cObjName, "" )
          ENDIF
        ENDIF  && 'object' $ lcLine
        
      NEXT              
      *** Return Final Status    
      RETURN llRetVal
    ENDWITH
  ENDFUNC && ChkMethodCode
  
  ********************************************************************
  *** [P] GETNEWCLASS: Check a line of code for Object instantiation
  ********************************************************************
  PROTECTED FUNCTION GetNewClass( tcLine )
  LOCAL loRetObj, lcObj, lcClass, lcLib
    WITH This
      *** Need a parameter object here
      loRetObj = NEWOBJECT( 'empty' )
      ADDPROPERTY( loRetObj, 'cClass', '' )
      ADDPROPERTY( loRetObj, 'cLib', '' )
      ADDPROPERTY( loRetObj, 'cBase', '' )
      ADDPROPERTY( loRetObj, 'cObjName', '' )
      
      *** Process it accordingly to how it is created
      *** Is this a method call?
      IF ".newobject" $ tcline OR ".addobject" $ tcline
        *** Object name is first, 
        lcObj = STREXTRACT( tcLine, "(", "," )
        *** Then class name is second
        lcClass = STREXTRACT( tcLine, ",", ",", 1, 2 )
        *** Classlibrary is third, if present!
        lcLib = STREXTRACT( tcLine, ",", ",", 2, 2 )
        *** If no library this must have been .AddObject()
        lcLib = IIF( EMPTY( lcLib ), ".AddObject Call", lcLib )
      ELSE
        *** Either NEWOBJECT() or CREATEOBJECT()
        IF 'newobject' $ tcLine OR 'createobject' $ tcLine    
          *** Object Name is word before "="
          lcObj = GETWORDNUM( tcLine, 1, "=" )
          *** Class name between "(" and first ","
          lcClass = STREXTRACT( tcLine, "(", "," )
          *** Classlibrary is next, if present!
          lcLib = STREXTRACT( tcLine, ",", ",", 1, 2 )
          *** If no library this must have been CreateObject()
          lcLib = IIF( EMPTY( lcLib ), "CreateObject Call", lcLib )
        ELSE
          *** not an object creation after all
          lcObj = ''
          lcClass = ''
          lcLib = ''
        ENDIF
      ENDIF
      
      *** Package up the results
      WITH loRetObj
        *** Might be a closing paren on the class name....
        .cClass   = ALLTRIM( CHRTRAN( lcClass, [)"']+CHR(9)+CHR(0), '' ))
        .cLib     = ALLTRIM( CHRTRAN( lcLib, [)"']+CHR(9)+CHR(0), '' ))
        .cObjName = ALLTRIM( CHRTRAN( lcObj, [)"']+CHR(9)+CHR(0), '' ))
      ENDWITH
      
      *** Return Final Status    
      RETURN loRetObj

    ENDWITH
  ENDFUNC && GetNewClass

  ********************************************************************
  *** [E] GetDirs: Build a directory tree and save it to the collection
  ********************************************************************
  FUNCTION GetDirs( tcStartDir )
  LOCAL ARRAY laDirs[1]
  LOCAL llRetVal, lnDirs, lnCnt, lcLastDir, lcFoundDir, lcKey, lcCurDir
    WITH This
      IF EMPTY( tcStartDir )
        *** Start at the specified root directory
        tcStartDir = .cRootDir
      ENDIF
      *** Save where we are
      lcLastDir = FULLPATH( CURDIR())

      *** Set the default directory
      SET DEFAULT TO (tcStartDir)

      *** Increment the directory counter and set the key
      .nLevels = .nLevels + 1
      lcKey = "DIR" + PADL( .nLevels, 3, '0' ) + "-"

      *** And save the current location
      lcCurDir = FULLPATH( CURDIR())
      
      *** Get the list of sub-directories
      lnDirs = ADIR( laDirs, '*.', 'D' ) 
      FOR lnCnt = 1 TO lnDirs
        lcFoundDir = laDirs[ lnCnt, 1]
        IF INLIST( lcFoundDir, '.', '..' )
          *** Ignore relative references to self and parent
          LOOP
        ELSE
          *** Make sure it really is a directory
          IF NOT DIRECTORY( lcFoundDir )
            LOOP
          ENDIF
        ENDIF
        *** Add this directory to the collection
        .oDirs.Add( lcCurDir + lcFoundDir, lcKey + lcFoundDir )
        .GetDirs( lcCurDir + lcFoundDir )
      NEXT

      *** Restore where we were
      SET DEFAULT TO (lcLastDir)

      *** Return Final Status    
      RETURN llRetVal

    ENDWITH
  ENDFUNC && GetDirList

  ********************************************************************
  *** [E] SETSELF: Initialize the object
  ********************************************************************
  FUNCTION SetSelf( tcRootDir )
  LOCAL llRetVal
    WITH This
      TRY
        *** Starting location for the search
        .cRootDir = tcRootDir
        *** Current working directory
        .cHomeDir = FULLPATH( CURDIR())
        *** Collection to hold the directory list
        .oDirs = CREATEOBJECT( "collection" )
        *** Make sure we are in the target directory
        IF NOT ( FULLPATH( CURDIR() ) == FULLPATH( .cRootDir ))
          SET DEFAULT TO FULLPATH( .cRootDir )
        ENDIF
        
        *** Create the result cursor
        CREATE CURSOR cur_Classes ( ;
          cLocn VARCHAR(200), ;
          cFile VARCHAR(100), ;
          cClas VARCHAR(100), ;
          cCLib VARCHAR(100), ;
          cBase VARCHAR(100), ;
          cObj  VARCHAR(100), ;
          cParent varchar(100)) 
        
        *** If we get to here, return value is OK
        llRetVal = .T.
      CATCH TO loErr
        MESSAGEBOX( loErr.message, 16, 'Setup Error' )
      
      ENDTRY
      *** Return status
      RETURN llRetVal

    ENDWITH
  ENDFUNC && SetSelf
  
  ********************************************************************
  *** [E] DESTROY: Reset to original home directory
  ********************************************************************
  FUNCTION Destroy()
  LOCAL llRetVal
    WITH This
      SET DEFAULT TO FULLPATH( .cHomeDir )      
      *** Return Final Status    
      RETURN llRetVal
    ENDWITH
  ENDFUNC && Destroy

  ********************************************************************
  *** [E] INIT: Sets starting conditions by calling SetSelf()
  ********************************************************************
  FUNCTION Init( tcRootDir )
  LOCAL llRetVal
    WITH This
      *** Initialize the object here
      llRetVal = .SetSelf( tcRootDir )
      *** Return Final Status    
      RETURN llRetVal
    ENDWITH
  ENDFUNC && Init

ENDDEFINE
